#include "mazebacktracking.h"

int checkValidPos(int mazeArray[HEIGHT][WIDTH], MapPosition pos)
{
    if (pos.x >= 0 && pos.x < WIDTH && pos.y >= 0 && pos.y < HEIGHT) {
        if (mazeArray[pos.y][pos.x] != WALL && mazeArray[pos.y][pos.x] != VISIT) {
            return 1;
        }
    }
    return 0;
}

void maze_BackTracking_Iterative(int mazeArray[HEIGHT][WIDTH],
              MapPosition startPos,
              MapPosition endPos,
              LinkedStack *pStack)
{
    StackNode *pNode = NULL;
    int isEmpty = 0, isFound = 0;
    int markArray[HEIGHT][WIDTH] = {0,};
    
    if (pStack != NULL)	{
        MapPosition newPosition = startPos;
        newPosition.direction = 0;
        pushLS(pStack, newPosition);
        while(isEmpty == 0 && isFound == 0) {
            pNode = popLS(pStack);
            if (pNode != NULL) {
                int x = pNode->data.x;
                int y = pNode->data.y;
                int direction = pNode->data.direction;
                
                while(isFound == 0 && direction < NUM_DIRECTIONS) {
                    int newX = x + DIRECTION_OFFSETS[direction][0];
                    int newY = y + DIRECTION_OFFSETS[direction][1];
                    
                    if (newX >= 0 && newX < WIDTH
                        && newY >= 0 && newY < HEIGHT
                        && mazeArray[newY][newX] == NOT_VISIT
                        && markArray[newY][newX] == NOT_VISIT) {
                        MapPosition newPosition;
                        newPosition.x = x;
                        newPosition.y = y;
                        newPosition.direction = direction + 1;
                        pushLS(pStack, newPosition);
                        markArray[y][x] = VISIT;
                        
                        x = newX;
                        y = newY;
                        direction = 0;
                        
                        if (newX == endPos.x && newY == endPos.y) {
                            isFound = 1;
                            newPosition.x = newX;
                            newPosition.y = newY;
                            newPosition.direction = 0;
                            pushLS(pStack, newPosition);
                            markArray[newY][newX] = VISIT;
                        }
                    }								
                    else {
                        direction++;
                    }
                }//end-of-while
                free(pNode);
            }//end-of-if
            isEmpty = isLinkedStackEmpty(pStack);
        }//end-of-while
    }
}

void showPathByStack(int mazeArray[HEIGHT][WIDTH], LinkedStack *pStack)
{
    StackNode *pNode = NULL;
    int resultArray[HEIGHT][WIDTH] = {0,};
    int isEmpty = 0;
    
    int i, j;
    for(i = 0; i < HEIGHT; i++) {
        for(j = 0; j < WIDTH; j++) {
            resultArray[i][j] = mazeArray[i][j];
        }
    }
    
    if (pStack != NULL) {
        i = 0;
        while(isEmpty == 0) {
            pNode = popLS(pStack);
            if (pNode != NULL) {
                int x = pNode->data.x;
                int y = pNode->data.y;
                resultArray[y][x] = VISIT;
                
                if (i > 0) {
                    printf("<-");
                }
                printf("(%d,%d) ", x, y);
                
                i++;
                free(pNode);
            }
            isEmpty = isLinkedStackEmpty(pStack);
        }
        printf("\n");
        printMaze(resultArray);
    }
}

void maze_BackTracking(int mazeArray[HEIGHT][WIDTH],
                       MapPosition endPos,
                       MapPosition currentPos,
                       MapPosition *pPath,
                       int *pStep)
{
    int step = *pStep;
    mazeArray[currentPos.y][currentPos.x] = VISIT;
    pPath[step] = currentPos;
    
    if (currentPos.x == endPos.x && currentPos.y == endPos.x) {
        return;
    }
    else {
        for(int i = 0; i < NUM_DIRECTIONS; i++) {
            MapPosition newPos = {0};
            newPos.x = currentPos.x + DIRECTION_OFFSETS[i][0];
            newPos.y = currentPos.y + DIRECTION_OFFSETS[i][1];
            
            if (checkValidPos(mazeArray, newPos)) {
                *pStep = step + 1;
                maze_BackTracking(mazeArray, endPos, newPos, pPath, pStep);
            }
        }
    }
}

void printMaze(int mazeArray[HEIGHT][WIDTH])
{
    int i, j;
    for(i = 0; i < HEIGHT; i++) {
        for(j = 0; j < WIDTH; j++) {
            if (mazeArray[i][j] == WALL) {
                printf("*");
            }
            else if (mazeArray[i][j] == VISIT) {
                printf("O");
            }
            else {
                printf(" ");
            }
        }
        printf("\n");
    }
}

void showPath(int mazeArray[HEIGHT][WIDTH], MapPosition *pPath, int step)
{
    int resultArray[HEIGHT][WIDTH] = {0,};
    for(int i = 0; i < HEIGHT; i++) {
        for(int j = 0; j < WIDTH; j++) {
            if (mazeArray[i][j] == WALL) {
                resultArray[i][j] = WALL;
            }
        }
    }
    for(int i = 0; i <= step; i++) {
        int x = pPath[i].x;
        int y = pPath[i].y;
        
        if (i > 0)
            printf("->");
        printf("(%d,%d)", y, x);
        
        resultArray[y][x] = VISIT;
    }
    printf("\n");
    printMaze(resultArray);
}
